
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "cgdk.h"
#include "car.h"
#include "file.h"

typedef int (*Tfcmp)(const void*, const void*);

#define CAR_BUFFER_SIZE 16384

TCAR *CAR_open (const char *fileName)
{
  TCAR *car;
  long itemSize;
  int i, j;

  // header size check (pre-compile)
  if ((sizeof (TCAR_header) == CAR_HEADERSIZE) && (sizeof (TCAR_item) == CAR_ITEMSIZE))
    ;
  else
    return NULL;

  // car allocation
  car = (TCAR *) _alloc (sizeof (TCAR));
  if (car == NULL)
  {
    return NULL;
  }

  car->type = 0;
  car->curItem = 0;

  // emulation check
  if (!((fileName == NULL) || (strcmp (fileName, CAR_EMULATION) == 0)))
  {
    // open CAR file
    if ((car->file = fopen (fileName, "rb")) == NULL)
    {
      _free (car);
      return NULL;
    }

    // read header
    if (fread ((void *) &car->header, sizeof (TCAR_header), 1, car->file) != 1)
    {
      fclose (car->file);
      _free (car);
      return NULL;
    }

    setvbuf (car->file, NULL, _IOFBF, CAR_BUFFER_SIZE);
    _mallocated += CAR_BUFFER_SIZE-BUFSIZ;

    // id, description and version check
    if ((car->header.id != CAR_ID) || (car->header.version > CAR_VERSION)
       || (strcmp (car->header.desc, CAR_DESC) != 0))
    {
      fclose (car->file);
      _free (car);
      return NULL;
    }

    // files item size check
    itemSize = car->header.filesCount*sizeof (TCAR_item);

    // seek file for file items
    if (fseek (car->file, -itemSize, SEEK_END) != 0)
    {
      fclose (car->file);
      _free (car);
      return NULL;
    }

    // allocate items memory
    car->item = (TCAR_item *) _alloc (itemSize);
    if (car->item == NULL)
    {
      fclose (car->file);
      _free (car);
      return NULL;
    }

    // read file items
    if ((fread (car->item, sizeof (TCAR_item), car->header.filesCount, car->file) != car->header.filesCount))
    {
      fclose (car->file);
      _free (car->item);
      _free (car);
      return NULL;
    }

    // set file position to the first file
//    if (fseek (car->file, sizeof (TCAR_header), SEEK_SET) != 0)
//    {
//      fclose (car->file);
//    _free (car->item);
//      _free (car);
//      return NULL;
//    }
    car->fPos = -1; // sizeof (TCAR_header);
  }
  else
  {
    car->type |= CAR_TEMU;
    car->file = NULL;
    car->item = (TCAR_item *) _alloc (sizeof (TCAR_item));
    strcpy (car->header.label, "");
    if (car->item == NULL)
    {
      _free (car);
      return NULL;
    }
  }

  return car;
}

static int fcmp (const void* s, const void* d)
{
  char *_s = (char *) s;
  char *_d = (char *) d;
  return stricmp (_s, _d);
}


static int CAR_findFile (TCAR *car, const char *fileName)
{
  char *s;

  if (car == NULL)
    return -1;

  if (strcmpi (fileName, car->item [car->curItem].name) != 0)
  {
    s = (char *) bsearch (fileName, &car->item [0].name, car->header.filesCount, sizeof (TCAR_item), (Tfcmp) stricmp);
    if (s == NULL)
    {
      return -1;
    }
    return (s - (char *) car->item) / sizeof (TCAR_item);
  }
  else
  {
    return car->curItem;
  }
}

static void CAR_loadCacheN (TCAR *car, int pos)
{
  if (car->item [pos].type & CAR_FCACHE)
    return;

  car->item [pos].buf = (char *) _alloc (car->item [pos].len+1);

  if (car->item [pos].buf == NULL)
    return;

  if (car->fPos != car->item [pos].begin)
  {
    if (fseek (car->file, car->item [pos].begin, SEEK_SET) == -1)
    {
      _free (car->item [pos].buf);
      return;
    }
    car->fPos = car->item [pos].begin;
  }

  if (fread (car->item [pos].buf, 1, car->item [pos].len, car->file) != car->item [pos].len)
  {
    _free (car->item [pos].buf);
    return;
  }
  car->fPos += car->item [pos].len;

  car->item [pos].type |= CAR_FCACHE;
}

static void CAR_unloadCacheN (TCAR *car, int pos)
{
  int fpos;

  if ((car->item [pos].type & CAR_FCACHE) == 0)
    return;

  if (pos == car->curItem)
  {
    fpos = car->item [pos].begin + car->curPos;
    if (car->fPos != fpos)
    {
      if (fseek (car->file, fpos, SEEK_SET) == -1)
      {
        return;
      }
      car->fPos = pos;
    }
  }

  if (car->item [pos].buf)
    _free (car->item [pos].buf);
  car->item [pos].buf = NULL;
  car->item [pos].type &= (-1-CAR_FCACHE);
}



void CAR_loadCache (TCAR *car, const char *fileName)
{
  int pos;

  if (car == NULL)
    return;

  if (car->type & CAR_TEMU)
    return;

  pos = CAR_findFile (car, fileName);
  if (pos == -1)
    return;

  CAR_loadCacheN (car, pos);
}

void CAR_unloadCache (TCAR *car, const char *fileName)
{
  int pos;

  if (car == NULL)
    return;

  if (car->type & CAR_TEMU)
    return;

  pos = CAR_findFile (car, fileName);
  if (pos == -1)
    return;

  CAR_unloadCacheN (car, pos);
}

void CAR_loadCacheAll (TCAR *car)
{
  int i;

  if (car == NULL)
    return;

  if (car->type & CAR_TEMU)
    return;

  for (i = 0; i < car->header.filesCount; i++)
  {
    CAR_loadCacheN (car, i);
  }
}

void CAR_unloadCacheAll (TCAR *car)
{
  int i;

  if (car == NULL)
    return;

  if (car->type & CAR_TEMU)
    return;

  for (i = 0; i < car->header.filesCount; i++)
  {
    CAR_unloadCacheN (car, i);
  }
}





void CAR_close (TCAR *car)
{
  int i;

  if (car != NULL)
  {
    if (car->file != NULL)
    {
      fclose (car->file);
      car->file = NULL;

      if ((car->type & CAR_TEMU) == 0)
      {
        _mallocated -= (CAR_BUFFER_SIZE-BUFSIZ);
      }
    }

    CAR_unloadCacheAll (car);

    if (car->item != NULL)
    {
      _free (car->item);
      car->item = NULL;
    }
    _free (car);
  }
}



// returns remaining length
long CAR_seek (TCAR *car, const char *fileName, long offset)
{
  char *s;
  long pos;

  if (car == NULL)
    return -1;

  if ((car->type & CAR_TEMU) == 0)
  {
    if (strcmpi (fileName, car->item [car->curItem].name) != 0)
    {
      s = (char *) bsearch (fileName, &car->item [0].name, car->header.filesCount, sizeof (TCAR_item), fcmp);
      if (s == NULL)
      {
        return -1;
      }
      car->curItem = (s - (char *) car->item) / sizeof (TCAR_item);
    }

    if (offset >= car->item [car->curItem].len)
    {
      return -1;
    }

    pos = car->item [car->curItem].begin + offset;

    // check for memory cache
    if ((car->item [car->curItem].type & CAR_FCACHE) == 0)
    {
      // seek if different position
      if (car->fPos != pos)
      {
        if (fseek (car->file, pos, SEEK_SET) == -1)
        {
          return -1;
        }
        car->fPos = pos;
      }
    }
    car->curPos = offset;

    return car->item [car->curItem].len - offset;
  }
  else
  {
    if (strcmpi (fileName, car->header.label) != 0)
    {
      if (car->file != NULL)
      {
        strcpy (car->header.label, "");
        fclose (car->file);
        car->file = NULL;
      }
      car->file = fopen (fileName, "rb");
      if (car->file == NULL)
        return -1;

      if (fseek (car->file, 0, SEEK_END) != 0)
      {
        fclose (car->file);
        car->file = NULL;
        return -1;
      }

      car->fPos = car->item->len = ftell (car->file);

      if (car->item->len == -1)
      {
        fclose (car->file);
        car->file = NULL;
        return -1;
      }
      strcpy (car->header.label, fileName);
    }

    if (offset > car->item->len)
    {
      return -1;
    }

    if (car->fPos != offset)
    {
      if (fseek (car->file, offset, SEEK_SET) != 0)
      {
        fclose (car->file);
        car->file = NULL;
        return -1;
      }
      car->fPos = offset;
    }

    return car->item->len - offset;
  }
}



int CAR_read (TCAR *car, void *buf, unsigned len)
{
  long left;

  if (car == NULL)
    return -1;

  // test for opened car file
  if (car->file == NULL)
    return -1;

  // emulation check
  if ((car->type & CAR_TEMU) == 0)
  {
    left = car->item [car->curItem].len - car->curPos;

    // read data size adjust to remaining data
    if (len > left)
      len = left;

    if (!(car->item [car->curItem].type & CAR_FCACHE))
    {
      // read the data from the car file
      len = fread (buf, 1, len, car->file);
      car->fPos += len;
    }
    else
    {
      memcpy (buf, car->item [car->curItem].buf + car->curPos, len);
    }
    car->curPos += len;

  }
  else
  {
    len = fread (buf, 1, len, car->file);
    car->fPos += len;
  }
  return len;
}

void *CAR_load (TCAR *car, const char *fileName)
{
  void *buf;
  long len;

  len = CAR_seek (car, fileName, 0);
  if (len == -1)
    return NULL;

  buf = _alloc (len+1);

  if (buf == NULL)
    return NULL;

  if (CAR_read (car, buf, len) != len)
  {
    _free (buf);
    return NULL;
  }
  return buf;
}


int CAR_getc (TCAR *car)
{
  int result;
  long left;

  if (car == NULL)
  {
    return EOF;
  }

  // test for opened car file
  if (car->file == NULL)
    return -1;

  // emulation check
  if ((car->type & CAR_TEMU) == 0)
  {
    left = car->item [car->curItem].len - car->curPos;

    // read data size adjust to remaining data
    if (left == 0)
      return EOF;

    if (!(car->item [car->curItem].type & CAR_FCACHE))
    {
      // read the character from the car file
      result = fgetc (car->file);
      if (result != EOF)
      {
        car->fPos++;
        car->curPos++;
      }
    }
    else
    {
      result = car->item [car->curItem].buf [car->curPos];
      car->curPos++;
    }
  }
  else
  {
    result = fgetc (car->file);
    if (result != EOF)
    {
      car->fPos++;
    }
  }
  return result;
}

















#ifdef __TURBOC__
  #define BUFLEN 16384
#else
  #define BUFLEN 262144
#endif

static int CAR_copyFiles (TCAR *car, TF_list *fileList)
{
  int i;
  FILE *file;
  TCAR_item *item;
  void *buf;
  long carPos;
  int count;
  int exitCode;

  carPos = sizeof (TCAR_header);
  exitCode = 0;

  buf = _alloc (BUFLEN);
  if (buf)
  {
    item = (TCAR_item *) _calloc (fileList->max, sizeof (TCAR_item), NULL);
    if (item != NULL)
    {
      for (i = 0; i < fileList->max; i++)
      {
        file = fopen (fileList->names [i], "rb");
        if (file == NULL)
        {
          exitCode = -1;
          break;
        }

        strcpy (item [i].name, fileList->names [i]);
        item [i].begin = carPos;

        while (count = fread (buf, 1, BUFLEN, file), count != 0 )
        {
          if (fwrite (buf, 1, count, car->file) != count)
          {
            fclose (file);
            exitCode = -1;
            break;
          }
          carPos += count;
        }
        if (exitCode != 0)
        {
          break;
        }

        item [i].len = carPos - item [i].begin;
      }
      if (exitCode == 0)
      {
        if (fwrite (item, fileList->max, sizeof (TCAR_item), car->file) != sizeof (TCAR_item))
        {
          exitCode = -1;
        }
      }
      _free (item);
    }
    _free (buf);
  }
  else
  {
    exitCode = -1;
  }
  return exitCode;
}



int CAR_create (const char *carFileName, const char *files)
{
  TF_list *fileList;
  int exitCode;
  TCAR car;

  if ((car.file = fopen (carFileName, "wb")) != NULL)
  {
    fileList = F_listNew (files);
    if (fileList != NULL)
    {
      memset (&car.header, 0, sizeof (TCAR_header));
      strcpy (car.header.desc, CAR_DESC);
      car.header.id = CAR_ID;
      car.header.version = CAR_VERSION;
      car.header.filesCount = fileList->max;

      if (fwrite (&car.header, 1, sizeof (TCAR_header), car.file) == sizeof (TCAR_header))
      {
        if (CAR_copyFiles (&car, fileList) == 0)
        {
          if (fclose (car.file) == 0)
          {
            F_listDelete (fileList);
            return 0;
          }
          else
          {
            exitCode =  -1;
          }
        }
        else
        {
          exitCode = -1;
        }
      }
      else
      {
        exitCode = -1;
      }

      F_listDelete (fileList);
    }
    else
    {
      exitCode = -1;
    }
    fclose (car.file);
  }
  else
  {
    exitCode = -1;
  }

  return exitCode;
}

int CAR_extract (const char *carFileName, const char *fileName)
{
  TCAR *car;
  FILE *file;
  int exitCode;
  int size;
  void *buf;

  buf = _alloc (BUFLEN);
  if (buf)
  {
    if (car = CAR_open (carFileName), car != NULL)
    {
      if ((size = CAR_seek (car, fileName, 0)) != -1)
      {
        file = fopen (fileName, "wb");
        if (file != NULL)
        {
          do
          {
            size = CAR_read (car, buf, BUFLEN);
            if (fwrite (buf, 1, size, file) != size)
            {
              fclose (file);
              _free (buf);
              return -1;
            }
          }
          while (size != 0);
          fclose (file);
          exitCode = 0;
        }
        else
        {
          exitCode = -1;
        }
      }
      else
      {
        exitCode = -1;
      }
      CAR_close (car);
    }
    else
    {
      exitCode = -1;
    }
    _free (buf);
  }
  else
  {
    exitCode = -1;
  }
  return exitCode;
}
